#include "endian.hh"

#include <alsa/asoundlib.h>
#include <GL/gl.h>

#include <vector>
#include <string>
#include <algorithm>

#include <stdlib.h>
#include <dlfcn.h>
#include <stdio.h>
#include <math.h>

#include <set>
#include <map>

#include <assert.h>

#include "rgbtorgb.hh"
#include "nesvideos-piece.hh"
#include "newhash.h"

#include "quantize.hh"
#include "lanczos.hh"

#include <glib-object.h>
#include <gdk/gdkevents.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkmain.h>

/* These two come from pSX: */
extern void quick_load_state(int);
extern void quick_save_state(int);

namespace
{
    /* BEGIN SETTINGS -- CHANGE THESE IF YOU WANT*/
    
    const int TargetWidth  = 320;  // -1 = don't change
    const int TargetHeight = 240;  // -1 = don't change
    const int TargetBpp    = 12;   // -1 = don't change

    const double fps = 60.0;

    const int SamplesPerSecond = (int)(44100 / fps);
    
    /* END SETTINGS */

    class Init
    {
    public:
        Init()
        {
            LoggingEnabled = 1;
            NESVideoSetVideoCmd(getenv("VIDEOLOG"));
            
            const char* rerec = getenv("RERECMODE");
            if(rerec) NESVideoSetRerecordingMode(strtol(rerec, 0, 10));
        }
    } Init;
    
    const std::string GetGLenumName(GLenum t)
    {
        const char*q = "?";
        #define c(s) case s: q=#s;break;
        switch(t)
        {
c(GL_2_BYTES) c(GL_BGRA) c(GL_FLOAT) c(GL_UNSIGNED_SHORT_5_6_5) c(GL_BGR)
c(GL_3_BYTES) c(GL_BLUE) c(GL_GREEN) c(GL_UNSIGNED_SHORT_4_4_4_4_REV)
c(GL_4_BYTES) c(GL_BYTE) c(GL_SHORT) c(GL_UNSIGNED_SHORT_5_5_5_1) c(GL_INT) 
c(GL_RGBA) c(GL_DEPTH_COMPONENT) c(GL_RGB) c(GL_UNSIGNED_BYTE_3_3_2) 
c(GL_ALPHA) c(GL_UNSIGNED_SHORT) c(GL_UNSIGNED_INT_8_8_8_8_REV) c(GL_UNSIGNED_INT)
c(GL_RED) c(GL_DOUBLE) c(GL_UNSIGNED_BYTE) c(GL_UNSIGNED_SHORT_5_6_5_REV)
c(GL_UNSIGNED_BYTE_2_3_3_REV) c(GL_UNSIGNED_INT_8_8_8_8) c(GL_LUMINANCE_ALPHA)
c(GL_UNSIGNED_SHORT_1_5_5_5_REV) c(GL_UNSIGNED_INT_10_10_10_2) c(GL_COLOR_INDEX)
c(GL_UNSIGNED_INT_2_10_10_10_REV) c(GL_UNSIGNED_SHORT_4_4_4_4) c(GL_LUMINANCE)
        }
        #undef c
        char Buf[512];
        sprintf(Buf, "0x%X (%s)", (int)t, q);
        return Buf;
    }

    const snd_pcm_channel_area_t* audiodata = 0;

    long audframes = 0;
    int got_video = 0;

    void DoSendNextFrame(const void* data, int wid,int hei, int bpp)
    {
        static int old_wid=0, old_hei=0, old_bpp=0;
        
        /* This assumes that any conversion has already been done. */
        if(wid != old_wid || hei != old_hei || bpp != old_bpp)
        {
            fprintf(stderr, "SendFrame: New OutProps: reso=%dx%d, bpp=%d\n", wid,hei,bpp);

            if(old_wid != 0 || old_hei != 0)
            {
                NESVideoNextAVI();
            }
            
            old_wid = wid;
            old_hei = hei;
            old_bpp = bpp;
        }

        NESVideoLoggingVideo(data, wid,hei,
            (unsigned) ( (1<<24) * fps ),
            bpp);
    }
    
    struct ImageCache
    {
        std::vector<unsigned char> cachedata;
        newhash_t hash;
        int w,h;
        GLenum f,t;
    } cache;

    template<int gmask, int bshift, int bpp, typename In>
    void ProduceOutputImage_2byte(int wid,int hei, 
                     const std::vector<Triplet<In> >& in,
                     bool swap_red_blue)
    {
        std::vector<unsigned char> Result(wid * hei * 2);
        
        const int maxpos = wid*hei;
        int x=0, y=0;
        
        const int g = 1;
        int r = 0, b = 2;
        if(swap_red_blue) std::swap(r,b);
        
        for(int pos=0; pos<maxpos; ++pos)
        {
            const int o16 = (x+4*y)%16;
            
            *(unsigned short*)&Result[pos*2]
                = (Quantize4x4<0x1F> (o16, in[pos][r] )<<0)
                | (Quantize4x4<gmask>(o16, in[pos][g] )<<5)
                | (Quantize4x4<0x1F> (o16, in[pos][b] )<<bshift);

            if(++x >= wid) { x=0;  ++y; }
        }
        DoSendNextFrame(&Result[0], wid, hei, bpp);
        cache.cachedata.swap(Result);
    }

    template<int stride, int bpp, typename In>
    void ProduceOutputImage_4byte(int wid,int hei, 
                     const std::vector<Triplet<In> >& in,
                     bool swap_red_blue)
    {
        std::vector<unsigned char> Result(wid * hei * stride);

        const int maxpos = wid*hei;
        int x=0, y=0;
        
        const int g = 1;
        int r = 0, b = 2;
        if(swap_red_blue) std::swap(r,b);
        
        for(int pos=0; pos<maxpos; ++pos)
        {
            const int o16 = (x+4*y)%16;

            Result[pos*stride+r] = Quantize4x4<0xFF>(o16, in[pos][0]),
            Result[pos*stride+g] = Quantize4x4<0xFF>(o16, in[pos][1]),
            Result[pos*stride+b] = Quantize4x4<0xFF>(o16, in[pos][2]);
            
            if(++x >= wid) { x=0;  ++y; }
        }
        DoSendNextFrame(&Result[0], wid, hei, bpp);
        cache.cachedata.swap(Result);
    }

    template<typename In>
    void ProduceOutputImage_I420(int wid,int hei, 
                     const std::vector<Triplet<In> >& in,
                     bool swap_red_blue)
    {
        const int stride = 3;
        std::vector<unsigned char> rgb24(wid * hei * stride);

        /* Convert to 24 */
        const int maxpos = wid*hei;
        int x=0, y=0;
        
        const int g = 1;
        int r = 0, b = 2;
        if(swap_red_blue) std::swap(r,b);
        
        for(int pos=0; pos<maxpos; ++pos)
        {
            const int o16 = (x+4*y)%16;

            rgb24[pos*stride+r] = Quantize4x4<0xFF>(o16, in[pos][0]),
            rgb24[pos*stride+g] = Quantize4x4<0xFF>(o16, in[pos][1]),
            rgb24[pos*stride+b] = Quantize4x4<0xFF>(o16, in[pos][2]);
            
            if(++x >= wid) { x=0;  ++y; }
        }
        
        std::vector<unsigned char> Result(wid*hei * 3/2);
        
        Convert24To_I420Frame(&rgb24[0], &Result[0], wid*hei, wid);

        DoSendNextFrame(&Result[0], wid, hei, 12);
        cache.cachedata.swap(Result);
    }

    template<typename In>
    void ProduceOutputImage_YUY2(int wid,int hei, 
                     const std::vector<Triplet<In> >& in,
                     bool swap_red_blue)
    {
        const int stride = 3;
        std::vector<unsigned char> rgb24(wid * hei * stride);

        /* Convert to 24 */
        const int maxpos = wid*hei;
        int x=0, y=0;
        
        const int g = 1;
        int r = 0, b = 2;
        if(swap_red_blue) std::swap(r,b);
        
        for(int pos=0; pos<maxpos; ++pos)
        {
            const int o16 = (x+4*y)%16;

            rgb24[pos*stride+r] = Quantize4x4<0xFF>(o16, in[pos][0]),
            rgb24[pos*stride+g] = Quantize4x4<0xFF>(o16, in[pos][1]),
            rgb24[pos*stride+b] = Quantize4x4<0xFF>(o16, in[pos][2]);
            
            if(++x >= wid) { x=0;  ++y; }
        }
        
        std::vector<unsigned char> Result(wid*hei * 2);
        
        Convert24To_YUY2Frame(&rgb24[0], &Result[0], wid*hei, wid);

        DoSendNextFrame(&Result[0], wid, hei, 17);
        cache.cachedata.swap(Result);
    }


    void DecodeInputImage_2byte
        (const void* videodata, int wid, int hei,
         std::vector<Triplet<unsigned char> >& out,
         int rpos,
         int gpos,int gmax,
         int bpos)
    {
        const int maxpos = wid*hei;
        const unsigned short* __restrict__ shortptr = (const unsigned short* ) videodata;
        
        Triplet<unsigned> shifts(rpos, gpos, bpos);
        Triplet<unsigned> maxs(31, gmax, 31);

        #pragma omp parallel for
        for(int pos=0; pos<maxpos; ++pos)
        {
            unsigned short val = shortptr[pos];
            out[pos] = ((val >> shifts) & maxs) * 255 / maxs;
        }
    }

    void DecodeInputImage_4byte
        (const void* videodata, int wid, int hei,
         std::vector<Triplet<unsigned char> >& out)
    {
        const int maxpos = wid*hei;
        const unsigned* __restrict__ wordptr = (const unsigned*)videodata;
        
        Triplet<unsigned> shifts(0, 8, 16);
        
        #pragma omp parallel for
        for(int pos=0; pos<maxpos; ++pos)
        {
            unsigned int val = wordptr[pos];
            out[pos] = val >> shifts; // target is unsigned char so a &0xFF is not needed
        }
    }

    void ProcessFrameUnmodified(unsigned in_bpp, const void* videodata, int wid, int hei)
    {
        /* bpp matches, no blue/red swap, don't need scaling */
        //fprintf(stderr, "Sending verbatim\n");
        DoSendNextFrame(videodata, wid,hei, in_bpp);
    }

    template<int gmask, int bshift, int bpp>
    void ProcessFrameUnscaledRedBlueSwap_2byte
         (const void* videodata, int wid, int hei)
    {
        std::vector<unsigned char> Result(wid * hei * 2);

        for(int pos=0, y=0; y<hei; ++y)
            for(int x=0; x<wid; ++x, ++pos)
            {
                unsigned val = ((const unsigned short*)videodata)[pos];
                
                unsigned b = ((val >>    0) & 0x1F);
                unsigned g = ((val >>    5) & gmask);
                unsigned r = ((val >>bshift)& 0x1F);
                
                *(unsigned short*)&Result[pos*2]
                    = ((r) << 0)
                    | ((g) << 5)
                    | ((b) << bshift);
            }
        //fprintf(stderr, "Sending swapped\n");
        DoSendNextFrame(&Result[0], wid,hei, bpp);
    }

    template<int stride, int bpp>
    void ProcessFrameUnscaledRedBlueSwap_4byte
        (const void* videodata, int wid, int hei)
    {
        /* 32 bpp in/out, need blue/red swap, don't need scaling */
        std::vector<unsigned char> Result(wid * hei * stride);

        for(int pos=0, y=0; y<hei; ++y)
            for(int x=0; x<wid; ++x, ++pos)
            {
                unsigned val =((const unsigned int *)videodata)[pos];
                
                unsigned char r = val >> 16;
                unsigned char g = val >> 8;
                unsigned char b = val >> 0;
                
                Result[pos*stride+0] = b;
                Result[pos*stride+1] = g;
                Result[pos*stride+2] = r;
            }
        //fprintf(stderr, "Sending swapped\n");
        DoSendNextFrame(&Result[0], wid,hei, bpp);
    }
    

    void RecordVideo(const void* videodata, int wid, int hei, GLenum format, GLenum type)
    {
        if(!videodata) return;
        
        if(audframes < SamplesPerSecond && false)
        {
            fprintf(stderr, "RecordVideo: skipping frames (audframes=%ld)\n", audframes);
            return;
        }
        audframes -= SamplesPerSecond;
        got_video = 1;
        
        static int old_wid=0,old_hei=0;
        static GLenum old_format=GL_FALSE, old_type=GL_FALSE;
        if(wid != old_wid || hei != old_hei
        || format != old_format || type != old_type)
        {
            old_wid = wid;
            old_hei = hei;
            old_type = type;
            old_format = format;
            fprintf(stderr, "RecordVideo: New InProps: reso=%dx%d, type=%s, format=%s\n",
                wid,hei,
                GetGLenumName(type).c_str(),
                GetGLenumName(format).c_str());
        }
        
        static long old_audframes = 0;
        if(audframes != old_audframes)
        {
            fprintf(stderr, "RecordVideo: Audio delay changed: now %ld (audio is %+ld frames ahead video)\n",
                audframes,
                audframes / SamplesPerSecond);
            old_audframes = audframes;
        }
        

        int samp=1, rpos=0, gpos=0,gmax=0, bpos=0;
        bool swap_red_blue = false;
        switch(format)
        {
            case GL_RGB:  swap_red_blue = false; break;
            case GL_RGBA: swap_red_blue = false; break;
            case GL_BGR:  swap_red_blue = true; break;
            case GL_BGRA: swap_red_blue = true; break;
            fprintf(stderr, "Unknown OpenGL pixel format: %s\n", GetGLenumName(format).c_str());
        }
        
        int in_bpp=-1;

        const int out_wid = (TargetWidth==-1 ? wid : TargetWidth);
        const int out_hei = (TargetHeight==-1 ? hei : TargetHeight);
        int out_bpp = TargetBpp;
        
        if(out_bpp == 24 || out_bpp == 32)
        {
            /* For 24 and 32 bpp, rb swap is the default in nesvideos-piece */
            swap_red_blue = !swap_red_blue;
        }
        
        #define bi(dp, sp, rp, gp,gm, bp) in_bpp=dp, samp=sp, rpos=rp, gpos=gp,gmax=gm, bpos=bp
        switch(type)
        {
            case GL_UNSIGNED_SHORT_5_6_5:
                swap_red_blue = !swap_red_blue; //fallthrough
            case GL_UNSIGNED_SHORT_5_6_5_REV:
                if(out_bpp == -1) out_bpp = 16;
                if(out_wid == wid && out_hei == hei && out_bpp==16) { 
                    if(swap_red_blue) { ProcessFrameUnscaledRedBlueSwap_2byte<0x3F,11,16>(videodata,wid,hei); return; }
                                 else { ProcessFrameUnmodified(16, videodata,wid,hei); return; } }
                bi(16, 2,  0, 5,63,11);
                break;
            case GL_UNSIGNED_SHORT_5_5_5_1:
                if(out_bpp == -1) out_bpp = 15;
                bi(15, 2, 11, 6,31, 1);
                break;
            case GL_UNSIGNED_SHORT_1_5_5_5_REV:
                if(out_bpp == -1) out_bpp = 15;
                if(out_wid == wid && out_hei == hei && out_bpp==15) { 
                    if(swap_red_blue) { ProcessFrameUnscaledRedBlueSwap_2byte<0x1F,10,15>(videodata,wid,hei); return; }
                                 else { ProcessFrameUnmodified(15, videodata,wid,hei); return; } }
                bi(15, 2,  0, 5,31,10);
                break;
            case GL_UNSIGNED_BYTE:
                if(out_bpp == -1) { out_bpp = 32; swap_red_blue = !swap_red_blue; }
                if(out_wid == wid && out_hei == hei && out_bpp==32) { 
                    if(swap_red_blue) { ProcessFrameUnscaledRedBlueSwap_4byte<4,32>(videodata,wid,hei); return; }
                                 else { ProcessFrameUnmodified(32,videodata,wid,hei); return; } }
                in_bpp=32;
                samp  = 4;
                break;
            fprintf(stderr, "Unknown OpenGL pixel type: %s\n", GetGLenumName(type).c_str());
        }
        #undef bi

        /* The generic case */
        const int in_dim  =     wid *     hei;
        //const int out_dim = out_wid * out_hei;

        /* Because Lanczos scaling is very slow, we cache previous results... */
        /* Caching may do good because often, frames are repeated */
        const newhash_t hash = newhash_calc((const unsigned char*)videodata, wid*hei*samp);
        
        if(hash == cache.hash
        && wid == cache.w
        && hei == cache.h
        && format == cache.f
        && type == cache.t
        && !cache.cachedata.empty())
        {
            //fprintf(stderr, "Sending cached\n");
            DoSendNextFrame(&cache.cachedata[0], out_wid, out_hei, out_bpp);
            return;
        }
        cache.hash = hash;
        cache.w    = wid;
        cache.h    = hei;
        cache.f    = format;
        cache.t    = type;
        
        if(out_wid == wid && out_hei == hei)
        {
            // No scaling is required.
            // Check if we've got one of our highly optimized bpp-changing cases.
            if((in_bpp==15 || in_bpp==16) && (out_bpp==24 && out_bpp==32))
            {
                std::vector<unsigned char> Result(wid*hei*(out_bpp==24 ? 3 : 4));
                
                if(in_bpp==15)
                    if(out_bpp==24) Convert15To24Frame(videodata, &Result[0], wid*hei, swap_red_blue);
                    else            Convert15To32Frame(videodata, &Result[0], wid*hei, swap_red_blue);
                else
                    if(out_bpp==24) Convert16To24Frame(videodata, &Result[0], wid*hei, swap_red_blue);
                    else            Convert16To32Frame(videodata, &Result[0], wid*hei, swap_red_blue);
                DoSendNextFrame(&Result[0], wid, hei, out_bpp);
                cache.cachedata.swap(Result);
                return;
            }
            if(in_bpp == 32 && out_bpp == 24 && !swap_red_blue)
            {
                std::vector<unsigned char> Result(wid*hei*3);
                Convert32To24Frame(videodata, &Result[0], wid*hei);
                DoSendNextFrame(&Result[0], wid, hei, out_bpp);
                cache.cachedata.swap(Result);
                return;
            }
            if(out_bpp==12 /* && !swap_red_blue */)
            {
                std::vector<unsigned char> Result(wid*hei*3/2);
                switch(in_bpp)
                {
                    case 15: Convert15To_I420Frame(videodata, &Result[0], wid*hei, wid); break;
                    case 16: Convert16To_I420Frame(videodata, &Result[0], wid*hei, wid); break;
                    case 24: Convert24To_I420Frame(videodata, &Result[0], wid*hei, wid); break;
                    case 32: Convert32To_I420Frame(videodata, &Result[0], wid*hei, wid); break;
                }
                DoSendNextFrame(&Result[0], wid, hei, out_bpp);
                cache.cachedata.swap(Result);
                return;
            }
            if(out_bpp==17 /* && !swap_red_blue */)
            {
                std::vector<unsigned char> Result(wid*hei*2);
                switch(in_bpp)
                {
                    case 15: Convert15To_YUY2Frame(videodata, &Result[0], wid*hei, wid); break;
                    case 16: Convert16To_YUY2Frame(videodata, &Result[0], wid*hei, wid); break;
                    case 24: Convert24To_YUY2Frame(videodata, &Result[0], wid*hei, wid); break;
                    case 32: Convert32To_YUY2Frame(videodata, &Result[0], wid*hei, wid); break;
                }
                DoSendNextFrame(&Result[0], wid, hei, out_bpp);
                cache.cachedata.swap(Result);
                return;
            }
        }
        
        std::vector<Triplet<unsigned char> > src(in_dim);  // srcx * srcy
        switch(samp)
        {
            case 2: DecodeInputImage_2byte(videodata,wid,hei, src, rpos, gpos,gmax, bpos); break;
            case 4: DecodeInputImage_4byte(videodata,wid,hei, src); break;
        }
        
        if(out_wid == wid && out_hei == hei)
        {
            /* Don't need scaling? Send now and don't play with floats */
            //fprintf(stderr, "Sending unscaled\n");
            switch(out_bpp)
            {
                case 12: ProduceOutputImage_I420(wid,hei, src, swap_red_blue); break;
                case 17: ProduceOutputImage_YUY2(wid,hei, src, swap_red_blue); break;
                case 15: ProduceOutputImage_2byte<0x1F,10,15>(wid,hei, src, swap_red_blue); break;
                case 16: ProduceOutputImage_2byte<0x3F,11,16>(wid,hei, src, swap_red_blue); break;
                case 24: ProduceOutputImage_4byte<3,24>(wid,hei, src, swap_red_blue); break;
                case 32: ProduceOutputImage_4byte<4,32>(wid,hei, src, swap_red_blue); break;
            }
            return;
        }
        
        std::vector<Triplet<LanczosFloat> > tgt; // will be sized automatically by Scaler
        LanczosScale(wid,hei, out_wid,out_hei, src, tgt);
        
        swap_red_blue = !swap_red_blue;
        
        /* FIXME: Seems to need swap_red_blue here, however,
         * when enabled, we get FPE
         */
        
#if 0
        fprintf(stderr, "Sending %dx%d (in: %dx%d)\n",out_wid,out_hei, wid,hei);
#endif
        switch(out_bpp)
        {
            case 12: ProduceOutputImage_I420(out_wid,out_hei, tgt, swap_red_blue); break;
            case 17: ProduceOutputImage_YUY2(out_wid,out_hei, tgt, swap_red_blue); break;
            case 15: ProduceOutputImage_2byte<0x1F,10,15>(out_wid,out_hei, tgt, swap_red_blue); break;
            case 16: ProduceOutputImage_2byte<0x3F,11,16>(out_wid,out_hei, tgt, swap_red_blue); break;
            case 24: ProduceOutputImage_4byte<3,24>(out_wid,out_hei, tgt, swap_red_blue); break;
            case 32: ProduceOutputImage_4byte<4,32>(out_wid,out_hei, tgt, swap_red_blue); break;
        }
    }

    void RecordAudio(snd_pcm_uframes_t offset, snd_pcm_uframes_t frames)
    {
        if(!audiodata) return;
        
        //static FILE* fp = fopen("s.out", "wb");
        const unsigned char* ptr = (const unsigned char*) audiodata->addr;
        ptr += offset * (audiodata->step/8);
        //fwrite(ptr, audiodata->step/8, frames, fp);
        //fflush(fp);
        
        if(audframes && !got_video) return; // Give video time
        
        audframes += frames;
        /* Seems pSX always gives 735 frames (perfect 44100/60 ratio) */
        
        NESVideoLoggingAudio(ptr, 44100, 16, 2, frames);
    }
}

namespace
{
    typedef gboolean (*GtkEventHandler) (GtkWidget*, GdkEventKey*, gpointer);
    GtkEventHandler OrigKeyDownHandler = 0;
    GtkEventHandler OrigKeyUpHandler = 0;
    
    static const int EnterKey = 36;
    static const int FrameAdKey1 = 51;
    static const int FrameAdKey2 = 59;
    
    static volatile enum { NotPaused, Paused, FrameAdvance } PauseMode = NotPaused;
    
    static long FrameCounter = 0;
    static std::map<int,bool> HeldKeysNow;
    static std::map<int,bool> HeldKeysPrev;
    static GtkWidget* KeyWidget = 0;
    static GdkEventKey* EventTemplate = 0;
    
    static int SelectedState = 1;
    
    static std::string GetHeldKeysList()
    {
        return "";/*
        if(HeldKeys.empty()) return "";
        
        std::string result = "Held";
        { char Buf[64]; sprintf(Buf, "[%d]", HeldKeysNoticed);
          result += Buf; }
        
        for(std::set<int>::const_iterator
            i = HeldKeys.begin(); i != HeldKeys.end(); ++i)
        {
            char Buf[64];
            sprintf(Buf, "(%d)", *i);
            result += Buf;
        }
        return result;*/
    }
    
    static void ReleaseKeys(const std::map<int,bool>& HeldKeys)
    {
        for(std::map<int,bool>::const_reverse_iterator
            i = HeldKeys.rbegin();
            i != HeldKeys.rend(); ++i)
        {
            GdkEventKey& ev = *EventTemplate;
            ev.hardware_keycode = i->first;
            ev.state  = 16;
            fprintf(stderr, "(%ld)Sending Up %d\n", FrameCounter, ev.hardware_keycode);
            OrigKeyUpHandler(KeyWidget, &ev, 0);
        }
    }
    
    static gboolean NewKeyDownHandler(GtkWidget*w, GdkEventKey*ev, gpointer data)
    {
        fprintf(stderr, "(%p)KeyDown(%lu)State(%lu)%s\n",
            (void*)w,
            (unsigned long) ev->hardware_keycode,
            (unsigned long) ev->state,
            GetHeldKeysList().c_str());
        
        // 10..19 = 1..0 , state 8 = alt
        // 41 = F
        // 39 = S
        // 44 = J
        // 46 = L
        
        if(ev->hardware_keycode >= 10 && ev->hardware_keycode <= 19
        && ev->state & 8)
        {
            SelectedState = ev->hardware_keycode - 10 + 1;
            fprintf(stderr, "State %d selected\n", SelectedState);
            return true;
        }
        if(ev->hardware_keycode == 44) // J
        {
            quick_save_state(SelectedState);
            return true;
        }
        if(ev->hardware_keycode == 46) // L
        {
            quick_load_state(SelectedState);
            return true;
        }
        
        if(ev->hardware_keycode >= 67 && ev->hardware_keycode < 72)
        {
            quick_load_state(ev->hardware_keycode - 67 + 1);
            return true;
        }
        if(ev->hardware_keycode >= 72 && ev->hardware_keycode < 77)
        {
            quick_save_state(ev->hardware_keycode - 72 + 1);
            return true;
        }

        if(ev->hardware_keycode == EnterKey)
        {
            if(ev->state & 8)
            {
                goto Passthru;
            }
            switch(PauseMode)
            {
                case NotPaused:
                    PauseMode = Paused;
                    fprintf(stderr, "Paused\n");
                    break;
                default:
                    PauseMode = NotPaused;
                    fprintf(stderr, "Unpaused\n");
            }
        }
        else
        if(ev->hardware_keycode == FrameAdKey1
        || ev->hardware_keycode == FrameAdKey2)
        {
            fprintf(stderr, "Frame advance\n");
            
            PauseMode = FrameAdvance;
            
            for(std::map<int,bool>::iterator k,
                i = HeldKeysPrev.begin(); i != HeldKeysPrev.end(); i=k)
            {
                k=i; ++k;
                std::map<int,bool>::const_iterator j = HeldKeysNow.find(i->first);
                if(j != HeldKeysNow.end() && j->second)
                {
                    HeldKeysPrev.erase(i);
                }
            }
            
            ReleaseKeys(HeldKeysPrev);
            HeldKeysPrev = HeldKeysNow;

            for(std::map<int,bool>::iterator k,
                i = HeldKeysNow.begin(); i != HeldKeysNow.end(); i=k)
            {
                k=i; ++k;
                if(!i->second)
                {
                    HeldKeysNow.erase(i);
                }
            }
        }
        else
        {
            if(PauseMode != NotPaused)
                HeldKeysNow[ev->hardware_keycode] = true;
        }
    Passthru:
        fprintf(stderr, "(%ld)Sending Down %d\n", FrameCounter, ev->hardware_keycode);
        return OrigKeyDownHandler(w,ev,data);
    }
    
    static gboolean NewKeyUpHandler(GtkWidget*w, GdkEventKey*ev, gpointer data)
    {
        EventTemplate = ev;
        KeyWidget     = w;
        fprintf(stderr, "(%p)KeyUp(%lu)State(%lu)%s\n",
            (void*)w,
            (unsigned long) ev->hardware_keycode,
            (unsigned long) ev->state,
            GetHeldKeysList().c_str());

        /*
        if(PauseMode == Paused
        && ev->hardware_keycode >= 67 && ev->hardware_keycode <= 76) // F1..F10
        {
            HeldKeysNow[ev->hardware_keycode] = false;
            return true; // Remember these keys as if they were held down.
        }
        */
        if(ev->hardware_keycode >= 67 && ev->hardware_keycode <= 76) // F1..F10
        {
            // Already handled
            return true;
        }
        
        if(PauseMode == FrameAdvance)
        {
            HeldKeysNow[ev->hardware_keycode] = false;
            return true;
        }
        
        if(ev->hardware_keycode == EnterKey && PauseMode == NotPaused)
        {
            for(std::map<int,bool>::iterator k,
                i = HeldKeysPrev.begin(); i != HeldKeysPrev.end(); i=k)
            {
                k=i; ++k;
                std::map<int,bool>::const_iterator j = HeldKeysNow.find(i->first);
                if(j != HeldKeysNow.end() && j->second)
                {
                    HeldKeysPrev.erase(i);
                }
            }
            for(std::map<int,bool>::iterator k,
                i = HeldKeysNow.begin(); i != HeldKeysNow.end(); i=k)
            {
                k=i; ++k;
                if(i->second)
                {
                    HeldKeysNow.erase(i);
                }
            }

            ReleaseKeys(HeldKeysPrev);
            ReleaseKeys(HeldKeysNow);
            HeldKeysPrev.clear();
            HeldKeysNow.clear();
        }
        
        fprintf(stderr, "(%ld)Sending Up %d\n", FrameCounter, ev->hardware_keycode);
        return OrigKeyUpHandler(w,ev,data);
    }
}

extern "C" 
{
    snd_pcm_sframes_t snd_pcm_avail_update(snd_pcm_t *pcm)
    {
        static struct Get2   
        {
            void *f;
            Get2() : f(dlopen("libasound.so", RTLD_LAZY)) {}
            ~Get2() { dlclose(f); }
            void *operator() (const char *s) { return dlsym(f, s); }
        } caller;

        // Find the real function
        typedef snd_pcm_sframes_t (*old_t)(snd_pcm_t*);
        static old_t old = (old_t)caller("snd_pcm_avail_update");
        
        snd_pcm_sframes_t res = old(pcm);
        
        res = 16384 - SamplesPerSecond;
            // Ensure that pSX always writes 735 sample packets,
            // even if it means lying about the soundcard's state.
            // (735 = 44100/60)
        
        return res;
    }

    int snd_pcm_mmap_begin
        (snd_pcm_t *pcm,
         const snd_pcm_channel_area_t **areas,
         snd_pcm_uframes_t *offset,
         snd_pcm_uframes_t *frames)
    {
        static struct Get2   
        {
            void *f;
            Get2() : f(dlopen("libasound.so", RTLD_LAZY)) {}
            ~Get2() { dlclose(f); }
            void *operator() (const char *s) { return dlsym(f, s); }
        } caller;

        // Find the real function
        typedef int (*old_t)(snd_pcm_t*, const snd_pcm_channel_area_t**, snd_pcm_uframes_t*, snd_pcm_uframes_t*);
        static old_t old = (old_t)caller("snd_pcm_mmap_begin");
        
        int res = old(pcm,areas,offset,frames);
        /*
        fprintf(stderr, "mmap begin: areas={%p,%u,%u}, offset=%lu,frames=%lu, res=%d\n",
            (*areas)->addr,
            (*areas)->first,
            (*areas)->step,
            *offset,
            *frames,
            res);
        */
        audiodata = (*areas);
        
        return res;
    }

    snd_pcm_sframes_t snd_pcm_mmap_commit
        (snd_pcm_t* pcm, snd_pcm_uframes_t offset, snd_pcm_uframes_t frames)
    {
        static struct Get2   
        {
            void *f;
            Get2() : f(dlopen("libasound.so", RTLD_LAZY)) {}
            ~Get2() { dlclose(f); }
            void *operator() (const char *s) { return dlsym(f, s); }
        } caller;

        // Find the real function
        typedef snd_pcm_sframes_t (*old_t)(snd_pcm_t*, snd_pcm_uframes_t, snd_pcm_uframes_t);
        static old_t old = 0;
        if(!old) old = (old_t)caller("snd_pcm_mmap_commit");
        
        snd_pcm_sframes_t res = old(pcm,offset,frames);
        /*
        fprintf(stderr, "mmap commit: %lu,%lu = %lu\n",
            offset,frames, res);
        */
        RecordAudio(offset, frames);

        return res;
    }
    
    void glTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels)
    {
        static struct Get2   
        {
            void *f;
            Get2() : f(dlopen("libGL.so", RTLD_LAZY)) {}
            ~Get2() { dlclose(f); }
            void *operator() (const char *s) { return dlsym(f, s); }
        } caller;

        // Find the real function
        typedef void (*old_t)(GLenum,GLint,GLint,GLint,GLsizei,GLsizei,GLenum, GLenum type,const GLvoid *);
        static old_t old = 0;
        if(!old) old = (old_t)caller("glTexSubImage2D");
        /*
        fprintf(stderr, "2dS level(%d)offs(%d,%d)format(0x%X)type(0x%X): %d,%d (%p)\n",
            (int)level,          // for pSX, always 0
            xoffset,yoffset,
            (int)format,         // for pSX, seems to be GL_BGRA
            (int)type,           // for pSX, seems to be GL_UNSIGNED_SHORT_1_5_5_5_REV
            width,height,
            pixels);
        */
        old(target,level,xoffset,yoffset,width,height,format,type,pixels);

        LoggingEnabled = 2;
        RecordVideo(pixels, width, height, format, type);
        
        ++FrameCounter;
    ReCheckPause:
        switch(PauseMode)
        {
            case FrameAdvance:
                PauseMode = Paused;
                break;
            case Paused:
                //fprintf(stderr, "[%d]Paused\n", gettid());
                //usleep(100000);
                gtk_main_iteration_do(true);
                goto ReCheckPause;
                break;
            case NotPaused:
                break;
        }
        //HeldKeysNoticed += 1;
    }
    
#undef fopen
    FILE* fopen(const char* path, const char* mode)
    {
        static struct Get2   
        {
            void *f;
            Get2() : f(dlopen("libc.so.6", RTLD_LAZY)) {}
            ~Get2() { dlclose(f); }
            void *operator() (const char *s) { return dlsym(f, s); }
        } caller;

        // Find the real function
        typedef FILE* (*old_t)(const char*,const char*);
        static old_t old = 0;
        if(!old) old = (old_t)caller("fopen");
        
        const char* quicksavepos = strstr(path, ".pSX/saves/quicksave_");
        if(quicksavepos)
        {
            const char* quicksavename = quicksavepos + 4 + 1 + 5 + 1 + 9 + 1;
            if(!strcmp(mode, "rb"))
            {
                fprintf(stderr, "Loading %s\n", quicksavename);
                NESVideoRerecordingLoad(quicksavename);
            }
            else if(!strcmp(mode, "wb"))
            {
                fprintf(stderr, "Saving %s\n", quicksavename);
                NESVideoRerecordingSave(quicksavename);
            }
        }
        return old(path, mode);
    }
    
    gulong   g_signal_connect_data            (gpointer       instance,
                           const gchar   *detailed_signal,
                           GCallback      c_handler,
                           gpointer       data,
                           GClosureNotify     destroy_data,
                           GConnectFlags      connect_flags)
    {
        static struct Get2   
        {
            void *f;
            Get2() : f(dlopen("libglibmm-2.0.so", RTLD_LAZY)) {}
            ~Get2() { dlclose(f); }
            void *operator() (const char *s) { return dlsym(f, s); }
        } caller;

        // Find the real function
        typedef gulong (*old_t)(gpointer,const gchar*,GCallback,gpointer,GClosureNotify,GConnectFlags);
        static old_t old = 0;
        if(!old) old = (old_t)caller("g_signal_connect_data");
        
        if(!strcmp(detailed_signal, "key_press_event") /*&& !OrigKeyDownHandler*/)
            { OrigKeyDownHandler = (GtkEventHandler) c_handler;
              c_handler = G_CALLBACK(NewKeyDownHandler); }
        if(!strcmp(detailed_signal, "key_release_event") /*&& !OrigKeyUpHandler*/)
            { OrigKeyUpHandler = (GtkEventHandler) c_handler;
              c_handler = G_CALLBACK(NewKeyUpHandler); }
        
        fprintf(stderr, "g_signal_connect_data(%s,%p,%p)\n", detailed_signal,
            c_handler,data);
        
        gulong result = 
            old(instance,detailed_signal,c_handler,data,destroy_data,connect_flags);
        
        return result;
    }
}
